#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <string.h>

#include "rc5-codes.h"
#include "beam.h"
#include "snake.h"
#include "morse.h"
#include "parith-15.h"

snake_t snake;

static const snake_level_t SNAKE_LEVELS[] PROGMEM = 
{
    [0] = 
    {
        .start_len = 20,
        .delay = 3,
        .grow  = 10,
        .raster = (1 << 3)
    }
};

// Tastendruck/Kommando
static void onKey_snake   (snake_t*, uint8_t code);

// Einen weiteren Pixel zeichnen
static void onPixel_snake (snake_t*);

// Menu-Punkt "Snake" wird aktiviert
static void onEnter_snake (snake_t*);

// tempus fugit
static void onTimeChanged_snake (snake_t*, long);

// Menu-Punkt "Snake"
const menuitem_t menuitem_snake PROGMEM = 
{
    .onPixel       = onPixel_snake,
    .onEnter       = onEnter_snake,
    .onTimeChanged = onTimeChanged_snake,
    .onKey         = onKey_snake,
    .use_fifo      = 1
};

// Richtungen
// NORT ^ SUED = 2
// WEST ^ OST  = 2
// ++ dreht nach links  (math. pos.)
// -- dreht nach rechts (math. neg.)
#define SN_N 3 // Nord
#define SN_S 1 // Sued
#define SN_W 0 // West
#define SN_E 2 // Ost

// Zum Zeichnen von Kopf + Futter
static const int8_t snake_head_dxy[] PROGMEM = 
{
    -1, -1,    -1,  0,    -1,  1,     0,  1,     1,  1,
     1,  0,     1, -1,     0, -1,     0,  0
};

// Tastendruck
void onKey_snake (snake_t * s, uint8_t code)
{
    // Kollision: nix tun
    if (s->draw.bums.body)
        return;
        
    uint8_t dir = s->dir;
    
    if (RC5_LEFT  == code)   dir = SN_W;
    if (RC5_RIGHT == code)   dir = SN_E;
    if (RC5_UP2   == code)   dir = SN_N;
    if (RC5_DOWN2 == code)   dir = SN_S;
    if (RC5_UP    == code)   dir = SN_N;
    if (RC5_DOWN  == code)   dir = SN_S;
    if (RC5_11 == code)      s->level.delay --;
    if (RC5_12 == code)      s->level.delay ++;

    dir &= 3;

    // Keine 180-Wende: Neue Richtung erst spaeter verarbeiten
    if (dir != s->body.seg[s->body.start].dir)
        s->new_dir = dir;
}

void onTimeChanged_snake (snake_t * s, long tid __attribute__((unused)))
{
    // bei Kollision: nix tun
    if (!s->draw.bums.body)
        s->changed = 1;
}

// Schlange einen Pixel weiter
void move_snake (snake_t * s)
{
    // Kopf 1 weiter
    uint8_t x = s->head.x;
    uint8_t y = s->head.y;
    
    uint8_t dir  = s->dir;
    uint8_t mask = s->raster_mask;
    
    // Falls Kopf auf Rasterpunkt: ggf. neue Richtung uebernehmen
    if (s->new_dir != 0xff 
        && (x & mask) == 0 && (y & mask) == 0)
    {
        s->dir = dir = s->new_dir;
        s->new_dir = 0xff;
    }

    // Schritt in die Richtung
    if (dir & 1)    y += dir - 2;
    else            x += dir - 1;
    
    // Kollision mit Rand (Kreis mit Radius 110)
    {
        int8_t x0 = x-0x80;
        int8_t y0 = y-0x80;
        
        if (x0*x0 + y0*y0 > 110*110)
            s->draw.bums.body = 1;
    }
    
    s->head.x = x;
    s->head.y = y;

    if (x == s->food.x  &&  y == s->food.y)
    {
        // Mampf!
        morse (_E_);
        s->grow = s->level.grow;
        
        // Neues Futter
        s->food.x = 0x60 + (uint8_t) (prandom() & 0x3f & ~mask);
        s->food.y = 0x60 + (uint8_t) (prandom() & 0x3f & ~mask);
    }
    
    // Erstes und letztes Segment anpassen
    uint8_t start = s->body.start;
    uint8_t nseg  = s->body.nseg;
    
    uint8_t dirs = 2 ^ dir;
    
    if (dirs != s->body.seg[start].dir)
    {
        // Richtung wurde gewechselt -> ein neues Segment vorhaengen
        nseg++;
        
        // start-- (mod SEG_MAX)
        
        if (--start == 0xff)
            start = SEG_MAX-1;

        // Start-Index hat sich geaendert
        s->body.start = start;
        // Neues Segment initialisieren
        s->body.seg[start].len = 0;
        s->body.seg[start].dir = dirs;
    }
    
    // Erstes Segment 1 laenger
    s->body.seg[start].len += 1;

    // Noch wachsen?
    if (s->grow)
        s->grow--;
    else
    {
        // Nicht mehr wachsen: letztes Segment 1 kuerzer
        // Index des letzten Segments (mod SEG_MAX)
        uint8_t idx  = start + nseg-1;
        if (idx >= SEG_MAX)
            idx -= SEG_MAX;
            
        uint8_t len = s->body.seg[idx].len - 1;
        
        if (len)
            // Segment 1 kuerzer
            s->body.seg[idx].len = len;
        else
           // Segment ist leer --> letztes Segment entfernen
           nseg--;
    };
       
    s->body.nseg = nseg;
}

// Schwierigkeitsgrad einstellen
void snake_set_level (snake_t * s, uint8_t level)
{
    memcpy_P (& s->level, & SNAKE_LEVELS[level] , sizeof (snake_level_t));
    s->level.level = level;
    s->raster_mask = s->level.raster-1;
}

// Menu-Punkt "Snake" aktivieren
void onEnter_snake (snake_t * s)
{
    snake_set_level (s, 0);
    
    // Reset snake
    s->new_dir = 0xff;
    s->grow = 0;
    s->changed = 1;
    
    // Schlong hat 1 Segment
    s->body.start = 0;
    s->body.nseg  = 1;
    s->body.seg[0].len = s->level.start_len;
    s->dir = prandom() & 3;
    s->body.seg[0].dir = 2 ^ s->dir;

    // Reset Pixel-Routine
    s->draw.len  = 0;
    s->draw.done = 1;
    s->draw.bums.body = 0;

    // 4 zufaellige bytes (.head und .food, jeweils .x und .y)
    uint8_t * pxy = & s->p[0].x;
    
    for (uint8_t i=0; i < 4; i++)
        pxy[i] = 0x60 + (uint8_t) (prandom() & 0x3f & ~snake.raster_mask);
}

// 1 Pixel aufs Rohr
void onPixel_snake (snake_t * s)
{
    if (s->draw.done)
    {
        // Alles ist fertig gezeichnet:
        uint8_t sreg = SREG;
        cli();
        uint16_t ticks = beam.ticks;
        SREG = sreg;

        // Nichts tun bis ticks = 700
        // Damit bei wenigen zu zeichnenden Pixeln diese
        // nicht heller sind.
        if (ticks < 700)
        {
            if (s->draw.done == 1)
            {
                // Strahl austasten und ins Aus stellen
                beam_skip();
                beam_xy (255, 255);
                s->draw.done = 2;
            }
            return;
        }

        // Ticks wieder auf 0,
        cli();
        beam.ticks = 0;
        SREG = sreg;
        
        // Zeichnen beginnt von vorne
        s->draw.done = 0;
        s->draw.len  = 0;
        s->draw.elem = 0xff-SN_FOOD-1;
        
        // onTimeChanged wurde aufgerufen: Schlange 1 weiter
        if (s->changed)
        {
            s->changed = 0;
            move_snake (s);
        }
    }
    
    // Schlange (Kopf, Body), Futter, etc. zeichnen
    // .draw.elem = 
    // fe --> FOOD
    // ff --> HEAD
    // 0...nseg-1 --> Segmente des Koerpers

    uint16_t len = s->draw.len;
    uint8_t dir  = s->draw.dir;
    uint8_t x    = s->draw.x;
    uint8_t y    = s->draw.y;
    uint8_t elem  = s->draw.elem;
    
    if (len == 0)
    {
        // Elemenet fertig: naechstes Segment
        // Springt der Strahl?
        uint8_t skip = 0;
        
        if (++elem == s->body.nseg)
        {
            s->draw.done = 1;
            return beam_skip();
        }
            
        if (elem >= 0x80)
        {
            // Koordinaten fuer Kopf, Futter, ...
            uint8_t idx = ~elem;
            skip = 1;
            s->draw.x = s->p[idx].x;
            s->draw.y = s->p[idx].y;
        }
        
        s->draw.elem = elem;
        
        if (elem < 0x80)
        {
            // Laenge und Richtung des Koerper-Segments
            uint8_t idx = elem + s->body.start;
            
            if (idx >= SEG_MAX)
                idx -= SEG_MAX;
            
            len = s->body.seg[idx].len;
            dir = s->body.seg[idx].dir;
        }
        else if (elem == 0xff-SN_HEAD)
        {
            // Kopf
            len = 7;
            dir = s->dir;
        }
        else if (elem == 0xff-SN_FOOD)
        {
            // Futter
            len = 9;
        }
        
        s->draw.dir = dir;
        
        if (skip)
        {
            // Strahl austasten
            s->draw.len = len;
            return beam_skip();
        }
    }

    s->draw.len = --len;

    if (elem >= 0x80)
    {
        // Pixel von Kopf, Futter, ... zeichnen
        uint16_t dxy = pgm_read_word (& snake_head_dxy[len << 1]);
        
        uint8_t dx = dxy;
        uint8_t dy = dxy >> 8;
        
        if (dir < 2)            dx = -dx, dy = -dy;
        if (!(dir & 1))         dx ^= dy, dy ^= dx, dx ^= dy;
        
        x += dx;
        y += dy;
    }
    else
    {
        // Body zeichnen
        // 1 Schritt in Segment-Richtung fortschreiten
        if (dir & 1)    y += dir-2;
        else            x += dir-1;
            
        s->draw.x = x;
        s->draw.y = y;

        // Kopf beisst in den Body        
        if (x == s->head.x && y == s->head.y)
            s->draw.bums.body = 1;
    }

    // Punkt auf die Roehre    
    beam_xy (x, y);
}
